// You must use "gcc" (sparc-elf-gcc) as assembly compiler (insead of default sparc-elf-as) with "-c" option.
// Project -> Properties -> C/C++ Build -> Settings. Toolchain Settings -> Cross GCC Assembler
// field 'Command:'  "gcc".
// Cross GCC Assembler -> General field 'Assembler flags:' -c (do not link files).
// Othervise the pre-processor will not be invoked and "#include xxx.h" and "#define xxx xxx" in the
// assembly files (*.S) will be ignored.
//
// As default CodeMaster-LEON use "as" as assembly compiler (sparc-elf-as,
// if you set correct prefix "sparc-elf-")

// "-nostartfiles" must NOT be used as Cross GCC Linker option.

#define OCRAM_START   0xA0000000
#define RAM_START     0x40000000
//      SRAM or SDRAM

#define OCRAM_SIZE    0x00010000
#define RAM_SIZE      0x00200000
//                    0x00010000 -  64 KByte - On-ChipRAM
//                    0x00100000 -   1 MByte
//                    0x00200000 -   2 MByte - SRAM board
//                    0x00F00000 -  15 MByte
//                    0x01000000 -  16 MByte
//                    0x01800000 -  24 MByte
//                    0x08000000 - 128 MByte - SDRAM board
//                    0x10000000 - 256 MByte
//                    0x20000000 - 512 MByte

#define _FRAME_POINTER (OCRAM_START + OCRAM_SIZE - 0x20)
// Maximum expected stack size:
#define _STACK_SIZE    0x00001000
//                     0x00000800 -  2 KByte
//                     0x00001000 -  4 KByte
//                     0x00002000 -  8 KByte
//                     0x00004000 - 16 KByte

#ifndef  MCFG_WC_CTRL
  #define MCFG_WC_CTRL 0x80100008

  #define MASK_5_BITS 0x001F

  #define WC_CTRL_PRWC_OFFSET  0
  #define WC_CTRL_PWWC_OFFSET  5
  #define WC_CTRL_SRWC_OFFSET 10
  #define WC_CTRL_SWWC_OFFSET 15
  #define WC_CTRL_IRWC_OFFSET 20
  #define WC_CTRL_IWWC_OFFSET 25
  #define WC_CTRL_PRWC_MASK   MASK_5_BITS
  #define WC_CTRL_PWWC_MASK   MASK_5_BITS
  #define WC_CTRL_SRWC_MASK   MASK_5_BITS
  #define WC_CTRL_SWWC_MASK   MASK_5_BITS
  #define WC_CTRL_IRWC_MASK   MASK_5_BITS
  #define WC_CTRL_IWWC_MASK   MASK_5_BITS

  // TODO: to set correct WC_CTRL_VALUE (it depends of cpu frequency and memory timings).
  #define WC_CTRL_VALUE (                                \
    ((0x4 & WC_CTRL_PRWC_MASK) << WC_CTRL_PRWC_OFFSET) | \
    ((0x4 & WC_CTRL_PWWC_MASK) << WC_CTRL_PWWC_OFFSET) | \
    ((0x1 & WC_CTRL_SRWC_MASK) << WC_CTRL_SRWC_OFFSET) | \
    ((0x0 & WC_CTRL_SWWC_MASK) << WC_CTRL_SWWC_OFFSET) | \
    ((0x0 & WC_CTRL_IRWC_MASK) << WC_CTRL_IRWC_OFFSET) | \
    ((0x0 & WC_CTRL_IWWC_MASK) << WC_CTRL_IWWC_OFFSET) )
#endif

#ifndef  CPU_ID_ADDRESS
 #define CPU_ID_ADDRESS      0xFFFFFFF0
 #define CPU_1906VM016_R3_ID 0x00001040
 #define CPU_1906VM016_R4_ID 0x0B041040
#endif

#define LOAD_DATA_TO_REGISTER(address, register_id) \
  sethi %hi(address), %register_id;                 \
  ld [%register_id + %lo(address)], %register_id;

#define STORE_REGISTER(address, register_id, temp_register_id) \
  sethi %hi(address), %temp_register_id;                       \
  st %register_id, [%temp_register_id + %lo(address)];

#define STORE_VALUE(address, value, register_0_id, register_1_id) \
  set value, %register_0_id;                                      \
  STORE_REGISTER(address, register_0_id, register_1_id);


// "_cleanregs_custom_weak" subroutine replace default weak subroutine.
  .text
  .global _cleanregs_custom_weak
_cleanregs_custom_weak:

_lowlevel_memory_controller_configuration:
  ! Set hardcoded value to MCFG_WC_CTRL (1906VM016_R4 or 1906VM016_R3 only)
  LOAD_DATA_TO_REGISTER(CPU_ID_ADDRESS, g1)
  set CPU_1906VM016_R4_ID, %g2
  subcc %g1, %g2, %g0
  be 2f
   set CPU_1906VM016_R3_ID, %g2
  subcc %g1, %g2, %g0
  bne 1f
   nop
2:
  STORE_VALUE(MCFG_WC_CTRL, WC_CTRL_VALUE, g1, g2)
// TODO: add another memory controller initialization here (before lable '1')

1:


_lowlevel_caches_initialization:
// Caches way_1 (all except way_0) initialization for 1906VM016 processor (software FIX)
initialize_caches_tags:
    sta %g0, [%g0] 0x02     ! Turn off cache
    flush                   ! Clear VALID fields (rewrite all tags of instruction and data caches)

initialize_icache_data_except_way_0:
! Note 1: STA / LDA require that the processor have been in supervisor mode (S bit of %psr must be set)
! Note 2: Instruction Cache tag/data access (ASI 0x0C-0x0D) require that Instruction Cache state must be "disabled"
! Note 3: STA / LDA to ASI 0x0C-0x0F during flush lead to data_access_exception trap

    set 0x00000008, %g3
    lda [%g3] 0x02, %g3     ! read Instruction Cache Configuration register
    srl %g3, 24, %g1        ! Instruction Cache Ways  (000 - direct mapping, 001 - 2 ways, 010 - 3 ways, 011 - 4 ways)
    and %g1, 0x7, %g1       ! mask is used
    add %g1, 1, %g1
    sll %g1, 10, %g2        ! Ways * 1 KByte
    srl %g3, 20, %g1        ! Instruction Cache WaySize (2^^WSIZE KBytes/Way)
    and %g1, 0xF, %g1       ! mask is used
    sll %g2, %g1, %g2       ! %g2 - Instruction Cache size
    set 0x00000400, %g3     ! 1 KByte
    sll %g3, %g1, %g3       ! %g3 - Instruction Cache single Way size

    ! wait until flush will be done
    lda [%g0] 0x02, %g1
2:
    srl %g1, 14, %g1
    andcc %g1, 0x0003, %g0  ! Check IP and DP bits of Cache Configuration register
    bne 2b
     lda [%g0] 0x02, %g1

    ! main write cycle
    .align 32               ! The right way is to place cycle into one prefetch block (8 instruction)
1:
    sub %g2, 4, %g2
    andncc %g2, %g3, %g0    ! exit from cycle before addresses of Way 0 begins
    bne 1b
     sta %g0, [%g2] 0x0D    ! write to ASI 0x0D (Instruction Cache data)
    .align 4                ! Restore word alignment

! Enable only instruction cache
    set 0x00000003, %g1
    sta %g1, [%g0] 0x02

initialize_dcache_data_except_way_0:
! Note 1: STA / LDA require that the processor have been in supervisor mode (S bit of %psr must be set)
    set 0x0000000C, %g3
    lda [%g3] 0x02, %g3     ! read Data Cache Configuration register
    srl %g3, 24, %g1        ! Data Cache Ways  (000 - direct mapping, 001 - 2 ways, 010 - 3 ways, 011 - 4 ways)
    and %g1, 0x7, %g1       ! mask is used
    add %g1, 1, %g1
    sll %g1, 10, %g2        ! Ways * 1 KByte
    srl %g3, 20, %g1        ! Data Cache WaySize (2^^WSIZE KBytes/Way)
    and %g1, 0xF, %g1       ! mask is used
    sll %g2, %g1, %g2       ! %g2 - Data Cache size
    set 0x00000400, %g3     ! 1 KByte
    sll %g3, %g1, %g3       ! %g3 - Data Cache single Way size
    ! main write cycle
1:
    sub %g2, 4, %g2
    andncc %g2, %g3, %g0    ! exit from cycle before addresses of Way 0 begins
    bne 1b
     sta %g0, [%g2] 0x0F    ! write to ASI 0x0F (Data Cache data)

! Enable both caches and data cache snooping
    set 0x0080000F, %g1
    sta %g1, [%g0] 0x02


_lowlevel_set_fp_and_sp:
  ! Set stack and frame pointers.
  set  _FRAME_POINTER, %fp
  andn %fp, 0xF, %fp
  sub  %fp, 0x60, %sp
  
  ! If memory area have Hamming(39, 32) error code protection then
  ! access to uninitialize word (32-bit) will produce trap to each 8-, 16-bit.
  ! Clearing of expected stack memory space (guard from 'stb/sth %rd, [%fp + %rs2/imm13]'
  ! and 'ldb/ldh [%fp + %rs2/imm13], %rd' accesses):
  set _STACK_SIZE, %g2
  sub %fp, %g2, %g2
  clr %g1
1:
  std %g0, [%g2]
  subcc %fp, %g2, %g0
  bg 1b
   add %g2, 8, %g2

  retl
   nop
