/*
 * _cleanregs_custom_weak.c
 *
 *  Created on: 17 march 2020
 *      Author: Sukhanov Mikhail (suhanov@niiet.ru)
 */

// Warning: Some level of optimization ('-O1', '-O2', '-O3', '-Os') must be used because
//          _cleanregs_custom_weak must be compiled as leaf subroutine (without
//          instructions 'save', 'call', 'ret', 'restore'). This requirement realised by
//          using '__attribute__((optimize(3), flatten))' of GCC 4.4.2 and .

// Note: All low level initialization must be placed at _cleanregs_custom_weak as inline functions.
//       '__attribute__((always_inline))' is used for all inline functions of '_lowlevel_*.h'.
#include "_lowlevel_memory_controller_configuration.h"
#include "_lowlevel_caches_initialization.h"
#include "_lowlevel_set_fp_and_sp.h"

// https://gcc.gnu.org/onlinedocs/gcc-4.4.2/gcc/Function-Attributes.html
void _cleanregs_custom_weak() __attribute__((optimize(3), flatten));
void _cleanregs_custom_weak() {
  _lowlevel_memory_controller_configuration();
  _lowlevel_caches_initialization();
  _lowlevel_set_fp_and_sp();
}

// "_cleanregs_custom_weak" label is point to inject low-level user-defined code
// (see "sparc-elf-4.4.2-mingw\src\libgloss\sparc_leon\" files "locore_mvt.S" and "locore.S")
// Standard execution sequence:
//   <start> / <_trap_table>
//   jump to <_hardreset_mvt> which
//     clear %asr17.SVT bit
//   jump to <_hardreset> which
//     call <_hardreset_custom_weak> (possible mkprom init code here),
//     set <_trap_table> as trap base address,
//     call <_hardreset_custom_svt_weak> (give mkprom svt chance to reset tbr),
//     call <_cleanregs_custom_weak>
//     ...
