/*
 * _lowlevel_memory_controller_configuration.h
 *
 *  Created on: 17 march 2020
 *      Author: Sukhanov Mikhail (suhanov@niiet.ru)
 */

#ifndef   MCFG_ADDRESS
  #define MCFG_ADDRESS 0x80100000

  typedef struct {
    volatile unsigned int lvds_ctrl;        // 0x00
    volatile unsigned int ft_ctrl;          // 0x04
    volatile unsigned int wc_ctrl;          // 0x08
    volatile unsigned int mem_ctrl;         // 0x0C
    volatile unsigned int checkbits_xor;    // 0x10
    volatile unsigned int sdram_cfg_1;      // 0x14
    volatile unsigned int sdram_cfg_2;      // 0x18
    volatile unsigned int sdclk_ctrl;       // 0x1C 1906VM016 rev.4 only
    volatile unsigned int bypass_ctrl;      // 0x20 1906VM016 rev.4 only
    volatile unsigned int bypass_checkbits; // 0x24 1906VM016 rev.4 only
  } niiet_mcfg;

  #define MASK_5_BITS 0x001F

  #define WC_CTRL_PRWC_OFFSET  0
  #define WC_CTRL_PWWC_OFFSET  5
  #define WC_CTRL_SRWC_OFFSET 10
  #define WC_CTRL_SWWC_OFFSET 15
  #define WC_CTRL_IRWC_OFFSET 20
  #define WC_CTRL_IWWC_OFFSET 25
  #define WC_CTRL_PRWC_MASK   MASK_5_BITS
  #define WC_CTRL_PWWC_MASK   MASK_5_BITS
  #define WC_CTRL_SRWC_MASK   MASK_5_BITS
  #define WC_CTRL_SWWC_MASK   MASK_5_BITS
  #define WC_CTRL_IRWC_MASK   MASK_5_BITS
  #define WC_CTRL_IWWC_MASK   MASK_5_BITS

  // TODO: to set correct WC_CTRL_VALUE (it depends of cpu frequency and memory timings).
  #define WC_CTRL_VALUE (                                \
    ((0x4 & WC_CTRL_PRWC_MASK) << WC_CTRL_PRWC_OFFSET) | \
    ((0x4 & WC_CTRL_PWWC_MASK) << WC_CTRL_PWWC_OFFSET) | \
    ((0x1 & WC_CTRL_SRWC_MASK) << WC_CTRL_SRWC_OFFSET) | \
    ((0x0 & WC_CTRL_SWWC_MASK) << WC_CTRL_SWWC_OFFSET) | \
    ((0x0 & WC_CTRL_IRWC_MASK) << WC_CTRL_IRWC_OFFSET) | \
    ((0x0 & WC_CTRL_IWWC_MASK) << WC_CTRL_IWWC_OFFSET) )
#endif

#ifndef   CPU_ID_ADDRESS
  #define CPU_ID_ADDRESS      0xFFFFFFF0
  #define CPU_1906VM016_R3_ID 0x00001040
  #define CPU_1906VM016_R4_ID 0x0B041040
#endif

static inline void _lowlevel_memory_controller_configuration() __attribute__((always_inline));
static inline void _lowlevel_memory_controller_configuration() {
  unsigned int *cpu_id_pointer = (unsigned int *)CPU_ID_ADDRESS;
  niiet_mcfg *mcfg = (niiet_mcfg *)MCFG_ADDRESS;

  // TODO: to add another memory controller register initialization.
  if ((*cpu_id_pointer == CPU_1906VM016_R4_ID) || (*cpu_id_pointer == CPU_1906VM016_R3_ID)) {
    mcfg->wc_ctrl = WC_CTRL_VALUE;
  }
}
